@@ -5,7 +5,8 @@ GLUE (Graph-Linked Unified Embedding)
 try:
     from importlib.metadata import version
 except ModuleNotFoundError:
-    from importlib_metadata import version
+    from pkg_resources import get_distribution
+    version = lambda name: get_distribution(name).version
 
 from . import data, genomics, graph, models, num, plot
 from .utils import config, log
